# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

from astropy.io import fits
import pandas as pd
import numpy as np
import copy
from .confreader import ConfReader

class ComputeXYREF:
    """
    ComputeXYREF is a class to compute xyref before computing trace and wavelength of a grism. If xyref is known, user can skip this process.
    - xyd = (x,y) from associated direct image of the source
    - xydiff = (dx,dy) for the relative position shift between direct and grism images. For HST images, this can be computed as xydiff = f(POSTARG1_d,POSTARG2_d,SCALE_d,POSTARG1_g,POSTARG2_g,SCALE_g) = (POSTARG1_g,POSTARG2_g) / SCALE_g - (POSTARG1_d,POSTARG2_d) / SCALE_d for *_d and *_g for values of direct and grism images, respectively
    - xyoff = (x,y) = h(confile,beam) or user specifies (Note: xyoff can be used as an manual adjustment)
    Use method compute() to compute xyref, which is accessible through self.xyref.
    Use method save() to save ./savefolder/saveprefix_xyref.csv; container must be specified.
    #####
    For convenience, 
    - if xydiff = 'default', user must specify gfile and dfile, and a method will be called to compute xydiff.
        gfile and dfile = (filepath, extension) for grism and direct images respectively.
        We assume POSTARG1,POSTARG2 in zeroth extension header, and IDCSCALE in the specified extension header.
        These values will be used to compute xydiff.
        A fits image from HST observation (e.g., MAST archive) should be already satisfied this assumption.
    - if xyoff = 'default', user must specify confile and beam, and a method will be called to retrieve xyoff from the confile.
        confile = filepath to configuration file
        beam = beam order
        We assume aXe configuration file with 'XOFF_A' and 'YOFF_A' for the xyoff given beam 'A' (for example)
    """
    def __init__(self,xyd,xydiff,xyoff,confile=None,beam=None,gfile=None,dfile=None,container=None):
        self.container = container
        self.data = {'xyd':xyd,
                     'xydiff':xydiff,
                     'xyoff':xyoff,
                     'confile':confile,
                     'beam':beam,
                     'gfile':gfile,
                     'dfile':dfile,
                     'xyref':'None'
                    }
        if xyoff == 'default':
            if confile is None or beam is None:
                raise ValueError('confile and beam must be specified for xyoff = "default"')
            self.data['xyoff'] = self._xyoff()
        if xydiff == 'default':
            if gfile is None or dfile is None:
                raise ValueError('gfile and dfile must be specified for xydiff = "default"')
            self.data['xydiff'] = self._xydiff()
    def _xyoff(self):
        confile = self.data['confile']
        beam = self.data['beam']
        conf = ConfReader(confile,beam)
        self.conf = conf
        confbeam = self.conf.beam
        return (confbeam['XOFF_'+beam].astype(float)[0],confbeam['YOFF_'+beam].astype(float)[0])
    def _xydiff(self):
        gfile = self.data['gfile']
        dfile = self.data['dfile']
        tmpg = fits.open(gfile[0])
        tmpd = fits.open(dfile[0])
        self.postarg_g= np.array([tmpg[0].header['POSTARG1'],tmpg[0].header['POSTARG2'],tmpd[gfile[1]].header['IDCSCALE']])
        self.postarg_d= np.array([tmpd[0].header['POSTARG1'],tmpd[0].header['POSTARG2'],tmpd[dfile[1]].header['IDCSCALE']])
        POSTARG1_g,POSTARG2_g,SCALE_g = self.postarg_g
        POSTARG1_d,POSTARG2_d,SCALE_d = self.postarg_d
        xydiff = np.array([POSTARG1_g,POSTARG2_g]) / SCALE_g - np.array([POSTARG1_d,POSTARG2_d]) / SCALE_d
        return xydiff
    ##########
    ##########
    ##########
    def compute(self):
        xyd,xydiff,xyoff = self.data['xyd'],self.data['xydiff'],self.data['xyoff']
        self.data['xyref'] = xyd + xydiff + xyoff
    ##########
    ##########
    ##########
    def save(self):
        if self.container is None:
            raise ValueError('container must be specified for saving. See hstgrism.container.Container class')
        saveprefix = self.container.data['saveprefix']
        savefolder = self.container.data['savefolder']
        string = './{1}/{0}_xyref.csv'.format(saveprefix,savefolder)
        tmp = copy.deepcopy(self.data)
        tmp['confile'] = (tmp['confile'],None)
        tmp['beam'] = (tmp['beam'],None)
        tmp = pd.DataFrame.from_dict(tmp, orient='index').T
        tmp.to_csv(string) 
        print('Save {0}'.format(string))
    


# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

from astropy.io import fits
import os

def to_fullframe(cutdata,fullfile,bb0x,bb1x,bb0y,bb1y,savesuffix,container):
    """
    to_fullframe replaces cutdata back to fullfile given bbcorner (bb0x,bb1x,bb0y,bb1y).
    The new file is created to ./savefolder/saveprefix_savesuffix.fits, given container and savesuffix.
    """
    saveprefix = container.data['saveprefix']
    savefolder = container.data['savefolder']
    # make a copy of fullfile to fullfile_new
    string = './{0}/{1}_{2}.fits'.format(savefolder,saveprefix,savesuffix)
    fullfile_new = (string,fullfile[1])
    os.system('cp {0} {1}'.format(fullfile[0],fullfile_new[0]))
    print('Make a copy of {0} to {1}'.format(fullfile[0],fullfile_new[0]))
    # replace
    tmp = fits.open(fullfile_new[0])
    tmp[fullfile_new[1]].data[bb0y:bb1y,bb0x:bb1x] = cutdata
    # save
    tmp.writeto(fullfile_new[0],overwrite=True)
    


# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

import matplotlib.pyplot as plt
import numpy as np

def show_bbox(cutoutdata,objname='No_object_name',
              maskdata=None,
              params = {'figsize':(10,10),
                        'cmap_cutout':'viridis',
                        'minmax':(5.,90.),
                        'cmap_mask':'Greys',
                        'alpha_mask':0.4,
                        'title':'default',
                        'fontsize':12
                       },
              do_trace=False,xcut=None,ycut=None,ww=None,
              tparams = {'tcolor':'red',
                         'tls':':',
                         'tlw':4,
                         'talpha':0.4
                        },
              aparams = {'axpertick':100,
                         'acolor_marker':'red',
                         'amarker':'o',
                         'afontsize':10,
                         'arotation':90.,
                         'acolor_text':'red'
                        },
              save=False,container=None
             ):
    """
    show_bbox plots Cutout2D object.
    - if maskdata is specified, maskdata will be plotted. Use params to set the parameters.
    - if d_trace is True, trace will be plotted. Use tparams and aparams to set parameters.
    - xcut = xg - bb0x and xg = xh + xref, and vice versa for ycut
    - if save is True, ./savefolder/saveprefix_bbox.plotformat is saved, given container.
    """
    figsize = params['figsize']
    cmap_cutout = params['cmap_cutout']
    minmax = params['minmax']
    plt.plot(figsize=figsize)
    m = np.isfinite(cutoutdata)
    vmin,vmax = np.percentile(cutoutdata[m],minmax[0]),np.percentile(cutoutdata[m],minmax[1])
    plt.imshow(cutoutdata,origin='lower',cmap=cmap_cutout,vmin=vmin,vmax=vmax)
    if maskdata is not None:
        cmap_mask = params['cmap_mask']
        alpha_mask = params['alpha_mask']
        plt.imshow(maskdata,origin='lower',cmap=cmap_mask,vmin=0.5,alpha=alpha_mask)
    if do_trace:
        if xcut is None or ycut is None or ww is None:
            raise ValueError('xcut,ycut,ww must be specified to plot trace')
        tcolor = tparams['tcolor']
        tls = tparams['tls']
        tlw = tparams['tlw']
        talpha = tparams['talpha']
        axpertick = aparams['axpertick']
        acolor_marker = aparams['acolor_marker']
        amarker = aparams['amarker']
        afontsize = aparams['afontsize']
        arotation = aparams['arotation']
        acolor_text = aparams['acolor_text']
        plt.plot(xcut,ycut,color=tcolor,ls=tls,lw=tlw,alpha=talpha)
        for i,ii in enumerate(xcut):
            if (i in {0,len(xcut)-1}) or (np.mod(i,axpertick)==0):
                label = '{0}A'.format(int(ww[i]))
                plt.plot(xcut[i],ycut[i],color=acolor_marker,marker=amarker)
                plt.annotate(label,(xcut[i],ycut[i]),
                             textcoords='offset points',
                             xytext=(0,10),
                             ha='center',
                             fontsize=afontsize,
                             rotation=arotation,
                             color=acolor_text
                            )
    if params['title'] == 'default':
        string = objname
    else:
        string = params['title']
    plt.title(string,fontsize=params['fontsize'])
    plt.tight_layout()
    if save:
        if container is None:
            raise ValueError('container must be specified to save')
        saveprefix = container.data['saveprefix']
        savefolder = container.data['savefolder']
        plotformat = container.data['plotformat']
        string = './{0}/{1}_bbox.{2}'.format(savefolder,saveprefix,plotformat)
        plt.savefig(string,format=plotformat,bbox_inches='tight')
        print('Save {0}'.format(string))
    


# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

import matplotlib.pyplot as plt
import numpy as np

def show_sum1d(gdata,objname,mdata=None,
               do_zero=True,
               params = {'figsize':(10,10),
                         'cmap':'viridis',
                         'minmax':(5.,90.),
                         'fontsize':12,
                         'title':'default',
                         'xlabel':'default',
                         'ylabel':'default'
                        },
               save=False,container=None):
    """
    show_sum1d takes a 2D data (i.e., gdata), computes summation along x and y axes, and plots.
    - do_zero = True will plot a zero line for reference.
    - mdata = mask with True for being excluded (i.e., set to np.nan)
    Use method save() to save the plot as ./savefolder/saveprefix_sum1d.plotformat, given container.
    """
    figsize = params['figsize']
    fontsize = params['fontsize']
    minmax = params['minmax']
    cmap = params['cmap']
    if mdata is None:
        mdata = np.full_like(gdata,False,dtype=bool)
    tmpdata = gdata.copy()
    tmpdata[mdata] = np.nan
    fig = plt.figure(figsize=figsize)
    # 221
    ax = fig.add_subplot(2,2,1)
    m = np.isfinite(tmpdata)
    vmin,vmax = np.percentile(tmpdata[m],minmax[0]),np.percentile(tmpdata[m],minmax[1])
    ax.imshow(tmpdata,origin='lower',cmap=cmap,vmin=vmin,vmax=vmax)
    string_title = objname if params['title'] == 'default' else params['title']
    string_xlabel = 'pixX' if params['xlabel'] == 'default' else params['xlabel']
    string_ylabel = 'pixY' if params['ylabel'] == 'default' else params['ylabel']
    ax.set_title(string_title,fontsize=fontsize)
    ax.set_xlabel(string_xlabel,fontsize=fontsize)
    ax.set_ylabel(string_ylabel,fontsize=fontsize)
    # 222
    ax2 = fig.add_subplot(2,2,2,sharey=ax)
    ny,nx = tmpdata.shape
    ax2.plot(np.nansum(tmpdata,axis=1),np.arange(ny))
    # 223
    ax3 = fig.add_subplot(2,2,3,sharex=ax)
    ny,nx = tmpdata.shape
    ax3.plot(np.arange(nx),np.nansum(tmpdata,axis=0))
    # do_zero
    if do_zero:
        ax2.plot(np.full_like(np.arange(ny),0.,dtype=float),np.arange(ny),'r:')
        ax3.plot(np.arange(nx),np.full_like(np.arange(nx),0.,dtype=float))
    fig.tight_layout()
    if save:
        if container is None:
            raise ValueError('container must be specified to save')
        saveprefix = container.data['saveprefix']
        savefolder = container.data['savefolder']
        plotformat = container.data['plotformat']
        string = './{2}/{0}_sum1d.{1}'.format(saveprefix,plotformat,savefolder)
        fig.savefig(string,format=plotformat,bbox_inches='tight')
        print('Save {0}'.format(string))



# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

from hstgrism.grismapcorr import GrismApCorr
from scipy.interpolate import interp1d
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

class Contamination:
    """
    Contamination is a class handling computation for grism contamination.
    - Object = object of interest
    - Contaminate = another object in the FoV of Object and contaminating it
    - trace_object_csv = trace.csv of Object (generated by HSTGRISM flow)
    - trace_contaminate_csv = trace.csv of Contaminate
    - halfdy_object = a 1D integer array parallel to Object trace for extraction aperture halfdy width of Object (excluding trace; that is total width 2+halfdy_object + 1).
     - if halfdy_object = None, contamination region would not be computed.
     - if Object was analyzed with HSTGRISM flow, tbox.csv has halfdy_object.
    - instrument = string for available instrument list in hstgrism.grismapcorr.GrismApCorr.
     - if not None, self.compute() will grab aperture correction values (apcorr) from GrismApCorr table given instrument.
     - self.apcorr to access apcorr as a dict with key as halfdy in pixel unit and value as 1D array parallel to self.combine_df.ww_contaminate
       - Note: for ww_contaminate where it is None, we replaced it with ww_contaminate.min(). This would not cause any conflict at the end since we will place None back.
     - self.combine_df after running self.compute() will also have columns corresponding to apcorr_contaminate_in and apcorr_contaminate_out.
    - contaminate_spectrum_wavelength,contaminate_spectrum_flam, sensitivity_curve_wavelength, and sensitivity_curve_value must be supplied to continue computing Contaminate cps inside the Object aperture.
     - Each is a 1D array.
     - contaminate_spectrum and sensitivity_curve does not need to be parallel.
     - Use contaminate_spectrum_interp_rule and sensitivity_curve_interp_rule to control how to interpolate.
    - Set do_drzblot_apcorr = True, and input drzblot_apcorr_wavelength, drzblot_apcorr_value to incorporate drzblot apcorr.
     - Use drzblot_apcorr_interp_rule to control the interpolate.
     - Note: drzblot apcorr must be computed given the same halfdy_object if following HSTGRISM flow.
    - cps_contaminate_nan_to_zero = True if cps_contaminate output should be replace np.nan by zero.
    Use self.compute() to start computation after properly specified information.
    - if halfdy_object is not None, contamination region would be computed.
    Use self.save(container) to save outputs to ./savefolder/saveprefix_savesuffix.extension
    - savefolder, saveprefix is controlled by Container class (see hstgrism.container.Container)
    - savesuffix.extension is preset:
     - comtaminateTrace.csv = merge trace of Contaminate to Object trace frame, accessible by self.combine_df
    Use self.show(save,container) to show traces, extraction region of Object, and contamination region
    - if save=True, the plot is saved at ./savefolder/saveprefix_contamination.plotformat where savefolder, saveprefix, and plotformat are defined by container (see hstgrism.container.Container)
    """
    def __init__(self,trace_object_csv,trace_contaminate_csv,halfdy_object=None,instrument=None,
                 contaminate_spectrum_wavelength=None,contaminate_spectrum_flam=None,
                 sensitivity_curve_wavelength=None,sensitivity_curve_value=None,
                 do_drzblot_apcorr=False,drzblot_apcorr_ww=None,drzblot_apcorr_value=None,
                 cps_contaminate_nan_to_zero=True,
                 contaminate_spectrum_interp_rule = {'kind':'linear','bounds_error':False,'fill_value':np.nan},
                 sensitivity_curve_interp_rule = {'kind':'linear','bounds_error':False,'fill_value':0.},
                 drzblot_apcorr_interp_rule = {'kind':'nearest','bounds_error':False,'fill_value':'extrapolate'},
                ):
        self.trace_object = pd.read_csv(trace_object_csv)
        self.trace_contaminate = pd.read_csv(trace_contaminate_csv)
        self.halfdy_object = halfdy_object
        self.instrument = instrument
        self.contaminate_spectrum_wavelength = contaminate_spectrum_wavelength
        self.contaminate_spectrum_flam = contaminate_spectrum_flam
        self.sensitivity_curve_wavelength = sensitivity_curve_wavelength
        self.sensitivity_curve_value = sensitivity_curve_value
        self.do_drzblot_apcorr = do_drzblot_apcorr
        self.drzblot_apcorr_ww = drzblot_apcorr_ww
        self.drzblot_apcorr_value = drzblot_apcorr_value
        self.cps_contaminate_nan_to_zero = cps_contaminate_nan_to_zero
        self.contaminate_spectrum_interp_rule = contaminate_spectrum_interp_rule
        self.sensitivity_curve_interp_rule = sensitivity_curve_interp_rule
        self.drzblot_apcorr_interp_rule = drzblot_apcorr_interp_rule
    def compute(self):
        xg_object = np.ceil(self.trace_object.xh.values + self.trace_object.xyref[0]).astype(int)
        yg_object = np.ceil(self.trace_object.yh.values + self.trace_object.xyref[1]).astype(int)
        ww_object = self.trace_object.ww.values
        xg_contaminate = np.ceil(self.trace_contaminate.xh.values + self.trace_contaminate.xyref[0]).astype(int)
        yg_contaminate = np.ceil(self.trace_contaminate.yh.values + self.trace_contaminate.xyref[1]).astype(int)
        ww_contaminate = self.trace_contaminate.ww.values
        yg_contaminate_wrt_object = np.array([],dtype=object)
        ww_contaminate_wrt_object = np.array([],dtype=object)
        for ii,i in enumerate(xg_object):
            if i in xg_contaminate:
                m = np.argwhere(xg_contaminate == i).flatten()
                yg_contaminate_wrt_object = np.concatenate((yg_contaminate_wrt_object,yg_contaminate[m]))
                ww_contaminate_wrt_object = np.concatenate((ww_contaminate_wrt_object,ww_contaminate[m]))
            else:
                yg_contaminate_wrt_object = np.concatenate((yg_contaminate_wrt_object,[None]))
                ww_contaminate_wrt_object = np.concatenate((ww_contaminate_wrt_object,[None]))
        combine_df = {'xg_object':xg_object,'yg_object':yg_object,'ww_object':ww_object,'yg_contaminate':yg_contaminate_wrt_object,'ww_contaminate':ww_contaminate_wrt_object.astype(float)}
        combine_df = pd.DataFrame(combine_df)
        self.combine_df = combine_df  
        ##### make halfdy_contaminate_in and out #####
        sentinel_halfdy_contaminate = False
        if self.halfdy_object is not None:
            self._configuration()
            self._halfdy_contaminate()
            sentinel_halfdy_contaminate = True
        else:
            print('halfdy_object must be specified when instantiate Contamination class in order to compute halfdy_contaminate')
        ##### prepare apcorr #####
        sentinel_apcorr_contaminate = False
        if self.instrument is not None and sentinel_halfdy_contaminate:
            self._apcorr_contaminate()
            sentinel_apcorr_contaminate = True
        else:
            print('instrument and halfdy_object must be specified when instantiate Contamination class in order to compute apcorr_contaminate')
        ##### compute cps_contaminate #####
        sentinel_cps_contaminate = False
        if self.contaminate_spectrum_wavelength is not None and self.contaminate_spectrum_flam is not None and self.sensitivity_curve_wavelength is not None and self.sensitivity_curve_value is not None and sentinel_halfdy_contaminate and sentinel_apcorr_contaminate:
            self._cps_contaminate()
            sentinel_cps_contaminate = True
        else:
            print('contaminate_spectrum, sensitivity_curve, instrument, and halfdy_object must be specified when instantiate Contamination class in order to compute cps_contaminate')
    def _configuration(self):
        ##### description for configuration #####
        # Given object_trace, contaminate_trace, and halfdy_object, there are three possible configurations needed to be treated differently:
        #  - '0': when contamination_trace == one of object_aperture = [object_trace - halfdy_object, object_trace + halfdy_object]
        #    - In this case, contamination is defined by only one bound of object_aperture at the far side.
        #  - '1': when contamination_trace is inside object_aperture = [object_trace - halfdy_object, object_trace + halfdy_object]
        #    - In this case, both object_aperture bounds determine contamination. Both bounds must be inclusive.
        #  - '2': when outside
        #    - In this case, closer bound is exclusive, but farther bound is inclusive.
        tyc = self.combine_df.yg_contaminate.copy()
        tx = self.combine_df.xg_object.values.copy()
        ty = self.combine_df.yg_object.values.copy()
        ty_up = ty + self.halfdy_object
        ty_low = ty - self.halfdy_object
        configuration = []
        for ii,i in enumerate(tyc):
            if tyc[ii] is None:
                configuration.append(None)
            else:
                if tyc[ii] == ty_low[ii] | tyc[ii] == ty_up[ii]:
                    configuration.append('0')
                elif tyc[ii] in np.arange(ty_low[ii],ty_up[ii]+1):
                    configuration.append('1')
                else:
                    configuration.append('2')  
        self.combine_df['configuration'] = configuration
    def _cps_contaminate(self):
        contaminate_spectrum_model = interp1d(self.contaminate_spectrum_wavelength,self.contaminate_spectrum_flam,**self.contaminate_spectrum_interp_rule)
        sensitivity_curve_model = interp1d(self.sensitivity_curve_wavelength,self.sensitivity_curve_value,**self.sensitivity_curve_interp_rule)
        ##### wwperpix_contaminate #####
        xhdiff = np.diff(self.combine_df.xg_object.values)
        xhdiff = np.append(xhdiff,xhdiff[-1]) # add one last element to maintain the array dimension
        wwdiff = np.diff(self.combine_df.ww_contaminate.values)
        wwdiff = np.append(wwdiff,wwdiff[-1]) # add one last element to maintain the array dimension
        wwperpix_contaminate = wwdiff/xhdiff
        ##### sensitivity_contaminate #####
        sensitivity_contaminate = sensitivity_curve_model(self.combine_df.ww_contaminate.values)
        ##### flam projecting to GRB frame #####
        flam_contaminate = contaminate_spectrum_model(self.combine_df.ww_contaminate.values)                            
        ##### drzblot_apcorr_contaminate #####
        if self.do_drzblot_apcorr:
            drzblot_apcorr_model = interp1d(self.drzblot_apcorr_ww,self.drzblot_apcorr_value,**self.drzblot_apcorr_interp_rule)
            drzblot_apcorr_contaminate = drzblot_apcorr_model(self.combine_df.ww_contaminate.values)
        else:
            drzblot_apcorr_contaminate = np.full_like(self.combine_df.ww_contaminate.values,1.,dtype=float)                           
        ##### apcorr_contaminate #####
        configuration = self.combine_df.configuration.values.copy()
        apcorr_contaminate = []
        for ii,i in enumerate(configuration):
            if configuration[ii] is None:
                apcorr_contaminate.append(None)
                continue
            elif configuration[ii] == '0':
                t = self.combine_df.apcorr_contaminate_far.values[ii] - self.combine_df.apcorr_contaminate_close.values[ii]
                t *= 0.5 # one-side contamination
                t += self.apcorr[0][ii]
            elif configuration[ii] == '1':
                t = self.combine_df.apcorr_contaminate_far.values[ii] + self.combine_df.apcorr_contaminate_close.values[ii]
                t *= 0.5 # one-side contamination
            elif configuration[ii] == '2':
                t = self.combine_df.apcorr_contaminate_far.values[ii] - self.combine_df.apcorr_contaminate_close.values[ii]
                t *= 0.5 # one-side contamination
            apcorr_contaminate.append(t)
        apcorr_contaminate = np.array(apcorr_contaminate).astype(float)
        ##### cps_contaminate #####
        cps_contaminate = flam_contaminate * wwperpix_contaminate * sensitivity_contaminate * drzblot_apcorr_contaminate * apcorr_contaminate
            # fix nan to zero
        if self.cps_contaminate_nan_to_zero:
            m = np.argwhere(~np.isfinite(cps_contaminate)).flatten()
            cps_contaminate[m] = 0.
        ##### output #####
        self.combine_df['wwperpix_contaminate'] = wwperpix_contaminate
        self.combine_df['sensitivity_contaminate'] = sensitivity_contaminate
        self.combine_df['flam_contaminate'] = flam_contaminate
        self.combine_df['drzblot_apcorr_contaminate'] = drzblot_apcorr_contaminate
        self.combine_df['apcorr_contaminate'] = apcorr_contaminate
        self.combine_df['cps_contaminate'] = cps_contaminate
    def _halfdy_contaminate(self):
        configuration = self.combine_df.configuration.values.copy()
        yg_object = self.combine_df.yg_object.values.copy()
        halfdy_object = self.halfdy_object.copy()
        yg_object_low,yg_object_up = yg_object-halfdy_object,yg_object+halfdy_object
        yg_contaminate = self.combine_df.yg_contaminate.values.copy()
        halfdy_contaminate_far = []
        halfdy_contaminate_close = []
        for ii,i in enumerate(configuration):
            ##### read description at def _configuration(self) #####
            if configuration[ii] is None:
                halfdy_contaminate_far.append(None)
                halfdy_contaminate_close.append(None)
                continue
            ta = yg_object_low[ii] - yg_contaminate[ii]
            ta_sign,ta_value = np.sign(ta),np.abs(ta)
            tb = yg_object_up[ii] - yg_contaminate[ii]
            tb_sign,tb_value = np.sign(tb),np.abs(tb)
            far = max(ta_value,tb_value)
            close = None
            if far == ta_value:
                s_far = ta_sign
                close = tb_value
                s_close = tb_sign
            elif far == tb_value:
                s_far = tb_sign
                close = ta_value
                s_close = ta_sign
            halfdy_contaminate_far.append(s_far*far)
            if configuration[ii] == '0':
                halfdy_contaminate_close.append(0)
            elif configuration[ii] == '1':
                halfdy_contaminate_close.append(s_close*close)
            elif configuration[ii] == '2':
                halfdy_contaminate_close.append((close-1)*s_close)
        self.combine_df['halfdy_contaminate_far'] = halfdy_contaminate_far
        self.combine_df['halfdy_contaminate_close'] = halfdy_contaminate_close
    def _apcorr_contaminate(self):
        halfdy_list = np.abs(np.concatenate((self.combine_df.halfdy_contaminate_far.unique(),self.combine_df.halfdy_contaminate_close.unique())))
        apcorr = {}
        for i in halfdy_list:
            if i is None:
                continue
            if i in apcorr.keys():
                continue
            halfdy_contaminate = i
            apsizepix = halfdy_contaminate*2 + 1
            apsizepix = np.full_like(self.combine_df.ww_contaminate.values,apsizepix,dtype=float)
            apsizearcsec = apsizepix * GrismApCorr().table[self.instrument]['scale']
            tww = self.combine_df.ww_contaminate.values.astype(float).copy()
            tww[np.isnan(tww)] = tww[np.isfinite(tww)].min()
            apcorrobj = GrismApCorr(instrument=self.instrument,apsize=apsizearcsec,wave=tww,aptype='diameter',apunit='arcsec',waveunit='A')
            apcorrobj.compute()
            apcorr[i] = apcorrobj.data['apcorr'] 
        self.apcorr = apcorr   
        ##### prepare apcorr_contaminate_in and apcorr_contaminate_out #####
        ##### place them into combine_df #####
        configuration = self.combine_df.configuration.values.copy()
        apcorr_contaminate_close = []
        apcorr_contaminate_far = []
        for ii,i in enumerate(configuration):
            if configuration[ii] is None:
                apcorr_contaminate_close.append(None)
                apcorr_contaminate_far.append(None)
            else:
                apcorr_contaminate_close.append(self.apcorr[np.abs(self.combine_df.halfdy_contaminate_close.values[ii])][ii])
                apcorr_contaminate_far.append(self.apcorr[np.abs(self.combine_df.halfdy_contaminate_far.values[ii])][ii])
        self.combine_df['apcorr_contaminate_close'] = apcorr_contaminate_close
        self.combine_df['apcorr_contaminate_far'] = apcorr_contaminate_far
    ##########
    ##########
    ##########
    def save(self,container=None):
        if container is None:
            raise ValueError('container must be specified to save.')
        string = './{0}/{1}_contaminateTrace.csv'.format(container.data['savefolder'],container.data['saveprefix'])
        self.combine_df.to_csv(string)
        print('Save {0}'.format(string))
    ##########
    ##########
    ##########
    def show(self,figsize=(10,10),
             object_color='black',object_ls=':',object_marker='x',
             contaminate_color='red',contaminate_ls=':',contaminate_marker='x',
             fontsize=12,
             save=False,container=None,
            ):
        plt.figure(figsize=figsize)
        combine_df = self.combine_df
        halfdy_object = self.halfdy_object
        plt.plot(combine_df.xg_object,combine_df.yg_object,ls=object_ls,color=object_color,label='Object trace')
        plt.plot(combine_df.xg_object,combine_df.yg_object + halfdy_object,color=object_color,marker=object_marker,label='Object aperture (inclusive)')
        plt.plot(combine_df.xg_object,combine_df.yg_object - halfdy_object,color=object_color,marker=object_marker)
        plt.plot(combine_df.xg_object,combine_df.yg_contaminate,color=contaminate_color,ls=contaminate_ls,label='Contaminate trace')
        plt.plot(combine_df.xg_object,combine_df.yg_contaminate + combine_df.halfdy_contaminate_close,color=contaminate_color,marker=contaminate_marker,label='Contaminate aperture (inclusive)')
        plt.plot(combine_df.xg_object,combine_df.yg_contaminate + combine_df.halfdy_contaminate_far,color=contaminate_color,marker=contaminate_marker)
        plt.xlabel('pixX',fontsize=fontsize)
        plt.ylabel('pixY',fontsize=fontsize)
        plt.title('Contamination',fontsize=fontsize)
        plt.legend(loc=(1.01,0.))
        plt.tight_layout()
        if save:
            if container is None:
                raise ValueError('container must be specified to save')
            string = './{0}/{1}_contamination.{2}'.format(container.data['savefolder'],container.data['saveprefix'],container.data['plotformat'])
            plt.savefig(string,plotformat=container.data['plotformat'],bbox_inches='tight')
            print('Save {0}'.format(string))          
            


# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

from polynomial2d.polynomial2d import Polynomial2D

class ComputeSIP:
    '''
    ComputeSIP compute SIP given (x1,x2) as xyref, and coef2d from ConfReader. Note: x1 is the leading term in SIP.
    Polynomial2D is used (pip install polynomial2d).
    Use method compress() to compute the 1D coefficients using SIP model, which is accessible through self.coef1d.
    '''
    def __init__(self,coef2d,xyref):
        self.coef2d = coef2d
        x1,x2 = xyref
        self.x1 = x1
        self.x2 = x2
    def compress(self):
        out = {}
        for i in self.coef2d:
            tmp = self.coef2d[i]   
            obj = Polynomial2D()
            obj.model['NORDER'] = tmp['NORDER']
            obj.model['COEF'] = tmp['COEF']
            obj.data['X1'] = self.x1
            obj.data['X2'] = self.x2
            obj.compute(rescale=False)
            out[i] = obj.model['YFIT']
        self.coef1d = out           



# Kornpob Bhirombhakdi
# kbhirombhakdi@stsci.edu

import os,glob,copy
from astropy.io import fits
from scipy.interpolate import interp1d
from scipy.signal import savgol_filter
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import json
from .background import Background
from .grismapcorr import GrismApCorr

class Extract:
    """
    This class was used in hstgrism version 7, and won't be used in any newer version.
    """
    def __init__(self,container,grismfile,sensefile,halfdy,
                 bkgfile=None,flatfile=None,tracefile=None,
                 objname='None',
                 params = {'padyup':5,'padylow':5,'padxleft':15,'padxright':15}
                ):
        if bkgfile is None:
            string = './{0}/{1}_bkg.fits'.format(container.data['savefolder'],container.data['saveprefix'])
            bkgfile = (string,1)
        if flatfile is None:
            string = './{0}/{1}_flat.fits'.format(container.data['savefolder'],container.data['saveprefix'])
            flatfile = (string,0)
        if tracefile is None:
            string = './{0}/{1}_trace.csv'.format(container.data['savefolder'],container.data['saveprefix'])
            tracefile = copy.deepcopy(string)
        self.container = container
        self.data = {'objname':objname,
                     'grismfile':grismfile,
                     'bkgfile':bkgfile,
                     'flatfile':flatfile,
                     'tracefile':tracefile,
                     'sensefile':sensefile,
                     'halfdy':halfdy,
                     'padyup':params['padyup'],
                     'padylow':params['padylow'],
                     'padxleft':params['padxleft'],
                     'padxright':params['padxright']
                    }
        self.grism = Background(container=container,objname=objname,gfile=grismfile,tfile=tracefile,
                                params_bbox = {'padxleft':params['padxleft'],'padxright':params['padxright'],'padyup':params['padyup'],'halfdyup':halfdy,'halfdylow':halfdy,'padylow':params['padylow'],'adjustx':0,'adjusty':0},
                               )
        self.bkg = Background(container=container,objname=objname,gfile=bkgfile,tfile=tracefile,
                              params_bbox = {'padxleft':params['padxleft'],'padxright':params['padxright'],'padyup':params['padyup'],'halfdyup':halfdy,'halfdylow':halfdy,'padylow':params['padylow'],'adjustx':0,'adjusty':0},
                               )
        self.flatfield = Background(container=container,objname=objname,gfile=flatfile,tfile=tracefile,
                                    params_bbox = {'padxleft':params['padxleft'],'padxright':params['padxright'],'padyup':params['padyup'],'halfdyup':halfdy,'halfdylow':halfdy,'padylow':params['padylow'],'adjustx':0,'adjusty':0},
                                   )
        self.sensitivity = self._get_sensitivity(sensefile)
    def _get_sensitivity(self,sensefile):
        tmp = fits.open(sensefile)
        return pd.DataFrame(tmp[1].data)
    ##########
    ##########
    ##########
    def save(self):
        saveprefix = self.container.data['saveprefix']
        savefolder = self.container.data['savefolder']
        saveformat = self.container.data['plotformat']
        # save self.data
        tmp = json.dumps(self.data)
        string = './{0}/{1}_extractdata.json'.format(savefolder,saveprefix)
        f = open(string,'w')
        f.write(tmp)
        f.close()
        print('Save {0}'.format(string))
        # save 2d images as fits: image, bkg, flat, mask
        primary_hdu = fits.PrimaryHDU()
        image_hdu = fits.ImageHDU()
        bkg_hdu = fits.ImageHDU()
        flat_hdu = fits.ImageHDU()
        mask_hdu= fits.ImageHDU()
        imagebkg_hdu = fits.ImageHDU()
        imagebkgflat_hdu = fits.ImageHDU()
        hdul = fits.HDUList([primary_hdu,image_hdu,bkg_hdu,flat_hdu,mask_hdu,imagebkg_hdu,imagebkgflat_hdu])
        hdul[1].header['EXTNAME'] = 'image'
        hdul[1].data = self.grism.bkg.data['Y']
        hdul[2].header['EXTNAME'] = 'bkg'
        hdul[2].data = self.bkg.bkg.data['Y']
        hdul[3].header['EXTNAME'] = 'flat'
        hdul[3].data = self.flatfield.bkg.data['Y']
        hdul[4].header['EXTNAME'] = 'mask'
        hdul[4].header['Description'] = 'value = 1 for the object'
        hdul[4].data = self.grism.bkg.data['MASK'].astype(int)
        hdul[5].header['EXTNAME'] = 'imbkg'
        hdul[5].header['Description'] = 'image - bkg'
        hdul[5].data = self.grism.bkg.data['Y'] - self.bkg.bkg.data['Y']
        hdul[6].header['EXTNAME'] = 'imbkgflat'
        hdul[6].header['Description'] = '(image - bkg)/flat'
        hdul[6].data = (self.grism.bkg.data['Y'] - self.bkg.bkg.data['Y'])/self.flatfield.bkg.data['Y']
        string = './{0}/{1}_extract.fits'.format(savefolder,saveprefix)
        hdul.writeto(string,overwrite=True)
        print('Save {0}'.format(string))
        # save trace as csv
        tmp = copy.deepcopy(self.extract['trace'])
        for i in {'XREF','YREF'}:
            tmpp = np.full_like(tmp['XG'],None,dtype=float)
            tmpp[0] = tmp[i]
            tmp[i] = tmpp.copy()
        for i in {'DYDX','DLDP'}:
            tmpp = np.full_like(tmp['XG'],None,dtype=float)
            for j,jj in enumerate(tmp[i]):
                tmpp[j] = tmp[i][j]
            tmp[i] = tmpp.copy()
        string = './{0}/{1}_extracttrace.csv'.format(savefolder,saveprefix)
        tmp = pd.DataFrame(tmp)
        tmp.to_csv(string)
        print('Save {0}'.format(string))
        # save bbox as csv
        tmp = copy.deepcopy(self.extract['bbox']['bbox'])
        for i in {'BBX','BBY'}:
            tmpp = np.full_like(tmp['XG'],None,dtype=float)
            for j,jj in enumerate(tmp[i]):
                tmpp[j] = tmp[i][j]
            tmp[i] = tmpp.copy()
        string = './{0}/{1}_extractbbox.csv'.format(savefolder,saveprefix)
        tmp = pd.DataFrame(tmp)
        tmp.to_csv(string)
        print('Save {0}'.format(string))
        # save 1d as csv
        tmp = copy.deepcopy(self.extract['1d'])
        for i in {'do_bkgsub','do_flat','do_apcorr'}:
            tmpp = np.full_like(tmp['cps'],None,dtype=float)
            tmpp[0] = self.extract[i]
            tmp[i] = tmpp.copy()
        for i in {'instrument'}:
            tmpp = np.full_like(tmp['cps'],None,dtype=object)
            tmpp[0] = self.extract[i]
            tmp[i] = tmpp.copy()
        string = './{0}/{1}_extract1d.csv'.format(savefolder,saveprefix)
        tmp = pd.DataFrame(tmp)
        tmp.to_csv(string)
        print('Save {0}'.format(string))        
    ##########
    ##########
    ##########
    def examine(self,save=True,
                params_extract={'central_rest':None,'width':None,'zshift':0.},
                params_savgol={'do_savgol':True,'window_length':5,'polyorder':0},
                plot_title=None,fontsize=12
               ):
        saveprefix = self.container.data['saveprefix']
        savefolder = self.container.data['savefolder']
        saveformat = self.container.data['plotformat']
        central_ww = params_extract['central_rest']
        width_ww = params_extract['width']
        zshift = params_extract['zshift']
        tmpww = self.extract['1d']['ww'].copy() / (1.+zshift)
        tmpflam = self.extract['1d']['flam'].copy()
        ww_bound = (central_ww - width_ww, central_ww + width_ww)
        m = np.argwhere((tmpww >= ww_bound[0])&
                        (tmpww <= ww_bound[1])).flatten()
        tmpx = tmpww[m]
        tmpy = tmpflam[m]
        plt.figure()
        plt.plot(tmpx,tmpy)
        if params_savgol['do_savgol']:
            window_length = params_savgol['window_length']
            polyorder = params_savgol['polyorder']
            plt.plot(tmpx,savgol_filter(tmpy,window_length=window_length,polyorder=polyorder))
        plt.xlabel('rest wavelength (A), zshift = {0}'.format(zshift),fontsize=fontsize)
        plt.ylabel('flam',fontsize=fontsize)
        if plot_title is None:
            plot_title = '{0}_{1}'.format(saveprefix,int(central_ww))
        plt.title(plot_title,fontsize=fontsize)
        plt.tight_layout()
        if save:
            string = './{0}/{1}_examine{3}.{2}'.format(savefolder,saveprefix,saveformat,int(central_ww))
            plt.savefig(string,format=saveformat,bbox_inches='tight')
            print('Save {0}'.format(string))
    ##########
    ##########
    ##########
    def compute(self,do_bkgsub=True,do_flat=True,do_apcorr=True,instrument=None,do_flam=True,fillvalue_apcorr='median'):
        data = self.grism.bkg.data['Y']
        bkg = self.bkg.bkg.data['Y']
        flat = self.flatfield.bkg.data['Y']
        mask = self.grism.bkg.data['MASK']
        trace = self.grism.trace
        bbox = self.grism.bbox
        ww = bbox['bbox']['WW']
        xg = bbox['bbox']['XG']
        yg = bbox['bbox']['YG']
        # we need
        # flam = cps / (apcorr * sensitivity * ww_per_pix)
        ##### cps
        if do_bkgsub:
            data = data - bkg
        if do_flat:
            data = data / flat
        cps = (data * mask.astype(float)).sum(axis=0)
        ##### ww_per_pix
        ww_per_pix = self._compute_wwperpix(ww,xg)
        ##### apcorr
        if do_apcorr:
            if instrument is None:
                string = 'instrument must be specified for do_apcorr = {0}'.format(do_apcorr)
                string += '\navailable instrument = {0}'.format(GrismApCorr().available_instrument)
                raise ValueError(string)
            apsize = mask.astype(float).sum(axis=0)
            grismapcorr = GrismApCorr(instrument=instrument,apsize=apsize,wave=ww,
                                      aptype='diameter',apunit='pix',waveunit='A'
                                     )  
            grismapcorr.compute(fill_value=fillvalue_apcorr)
            apcorr = grismapcorr.data['apcorr']
        else:
            apcorr = np.full_like(ww,1.,dtype=float)
        ##### sensitivity
        if do_flam:
            sensitivity_model = interp1d(self.sensitivity.WAVELENGTH,self.sensitivity.SENSITIVITY,kind='linear',bounds_error=False,fill_value=None)
            sensitivity = sensitivity_model(ww)
            sensitivity[~np.isfinite(sensitivity)] = 1. # replace nan with 1.
            flam = cps / (apcorr * sensitivity * ww_per_pix)
        else:
            sensitivity = np.full_like(ww,1.,dtype=float)
            flam = None
        ##### output
        self.extract = {'image':data,
                        'mask':mask,
                        'trace':trace,
                        'bbox':bbox,
                        'do_bkgsub':do_bkgsub,
                        'do_flat':do_flat,
                        'do_apcorr':do_apcorr,
                        'instrument':instrument,
                        'do_flam':do_flam,
                        '1d':{'cps':cps,'ww':ww,'xg':xg,'yg':yg,
                              'apcorr':apcorr,'sensitivity':sensitivity,'ww_per_pix':ww_per_pix,
                              'flam':flam
                             }
                       }
    def _compute_wwperpix(self,ww,xg):
        wwdiff = np.diff(ww)
        wwdiff = np.append(wwdiff,wwdiff[-1])
        xgdiff = np.diff(xg)
        xgdiff = np.append(xgdiff,xgdiff[-1])
        return wwdiff/xgdiff
    ##########
    ##########
    ##########
    def show(self,save=False,
             params={'figsize':(10,10),
                     'color':'red',
                     'ls':':',
                     'lw':4,
                     'marker':'x',
                     'alpha':0.2,
                     'fontsize':12,
                     'minmax_im':(5.,99.),
                     'xpertick_im':100,
                     'xpertick_1d':50,
                     'cmap_im':'viridis',
                     'cmap_mask':'Greys',
                     'annotate_level':0.,
                     'sensitivity_level':0.1e16,
                     'adjust_minmax_flam':(0.9,1.1)
             }
            ):
        saveprefix = self.container.data['saveprefix']
        savefolder = self.container.data['savefolder']
        saveformat = self.container.data['plotformat']
        adjust_minmax_flam = params['adjust_minmax_flam']
        minmax_im = params['minmax_im']
        xpertick_im = params['xpertick_im']
        xpertick_1d = params['xpertick_1d']
        color = params['color']
        ls = params['ls']
        alpha = params['alpha']
        lw = params['lw']
        marker = params['marker']
        fontsize = params['fontsize']
        figsize = params['figsize']
        cmap_im = params['cmap_im']
        cmap_mask = params['cmap_mask']
        annotate_level = params['annotate_level']
        sensitivity_level = params['sensitivity_level']
        image = self.extract['image']
        mask = self.extract['mask']
        xref = self.extract['trace']['XREF']
        yref = self.extract['trace']['YREF']
        trace_xg = self.extract['trace']['XG']
        trace_yg = self.extract['trace']['YG']
        trace_ww = self.extract['trace']['WW']
        bbox_bbx = self.extract['bbox']['bbox']['BBX']
        bbox_bby = self.extract['bbox']['bbox']['BBY']
        bbox_xg = self.extract['bbox']['bbox']['XG']
        bbox_ww = self.extract['bbox']['bbox']['WW']
        do_flam = self.extract['do_flam']
        cps = self.extract['1d']['cps']
        flam = self.extract['1d']['flam']
        sensitivity = self.extract['1d']['sensitivity']

        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(4,1,1)
        vmin,vmax = np.percentile(image,minmax_im[0]),np.percentile(image,minmax_im[1])
        ax.imshow(image,origin='lower',cmap=cmap_im,vmin=vmin,vmax=vmax)
        tmpx,tmpy = trace_xg,trace_yg
        bbx0,bby0 = bbox_bbx[0],bbox_bby[0]
        tmpx,tmpy = tmpx-bbx0,tmpy-bby0
        ww = trace_ww
        ax.plot(tmpx,tmpy,color=color,ls=ls,lw=lw,alpha=alpha)
        for i,ii in enumerate(tmpx):
            if (i in {0,len(tmpx)-1}) or (np.mod(i,xpertick_im)==0):
                label = '{0}A'.format(int(ww[i]))
                ax.plot(tmpx[i],tmpy[i],color=color,marker=marker)
                ax.annotate(label,(tmpx[i],tmpy[i]),
                             textcoords='offset points',
                             xytext=(0,10),
                             ha='center',
                             fontsize=fontsize,
                             rotation=0.,
                             color=color
                            )
        tmpmask = mask.astype(float)
        tmpmask[tmpmask==1.] = np.nan
        ax.imshow(tmpmask,origin='lower',cmap=cmap_mask,vmin=0.,alpha=alpha)
        string = '{0} {1} \ndo_bkgsub={2} do_flat={3}'.format(self.data['objname'],saveprefix,
                                                              self.extract['do_bkgsub'],self.extract['do_flat']
                                                             )
        ax.set_title(string,fontsize=fontsize)
        string = 'pixY - {0}'.format(yref)
        ax.set_ylabel(string,fontsize=fontsize)

        ax = fig.add_subplot(4,1,2,sharex=ax)
        tmpy = cps.copy()
        tmpx = bbox_xg.copy()
        bbx0 = bbox_bbx[0]
        tmpx = tmpx-bbx0
        ax.plot(tmpx,tmpy)
        ax.plot(tmpx,np.full_like(tmpx,annotate_level,dtype=float),color=color,ls=ls,lw=lw)
        tmpx2 = trace_xg - bbx0
        tmpy2 = np.full_like(tmpx2,annotate_level,dtype=float)
        ww = trace_ww
        for i,ii in enumerate(tmpx2):
            if (i in {0,len(tmpx2)-1}) or (np.mod(i,xpertick_1d)==0):
                label = '{0}A'.format(int(ww[i]))
                ax.plot(tmpx2[i],annotate_level,color=color,marker=marker)
                ax.annotate(label,(tmpx2[i],annotate_level),
                             textcoords='offset points',
                             xytext=(0,10),
                             ha='center',
                             fontsize=fontsize,
                             rotation=90.,
                             color=color
                            )
        ax.set_ylabel('cps',fontsize=fontsize)
        ax.grid()
        
        if do_flam:
            ax = fig.add_subplot(4,1,3,sharex=ax)
            tmpx = bbox_xg-bbx0
            tmpy = flam.copy()
            ax.plot(tmpx,tmpy)
            m = np.argwhere(sensitivity >= sensitivity_level).flatten()
            tmpx_good,tmpy_good,ww_good = tmpx[m],tmpy[m],bbox_ww[m]
            m = np.isfinite(tmpy_good)
            tmpx_good,tmpy_good,ww_good = tmpx_good[m],tmpy_good[m],ww_good[m]
            for i,ii in enumerate(tmpx_good):
                if (i in {0,len(tmpx_good)-1}) or (np.mod(i,xpertick_1d)==0):
                    label = '{0}A'.format(int(ww_good[i]))
                    ax.plot(tmpx_good[i],tmpy_good[i],color=color,marker=marker)
                    ax.annotate(label,(tmpx_good[i],tmpy_good[i]),
                                 textcoords='offset points',
                                 xytext=(5,-5),
                                 ha='left',
                                 fontsize=fontsize,
                                 rotation=0.,
                                 color=color
                                )
            ax.set_ylim(tmpy_good.min()*adjust_minmax_flam[0],tmpy_good.max()*adjust_minmax_flam[1])
            ax.set_ylabel('flam',fontsize=fontsize)
            ax.grid()

            ax = fig.add_subplot(4,1,4,sharex=ax)
            ax.plot(tmpx,sensitivity)
            ax.plot(tmpx,np.full_like(tmpx,sensitivity_level),ls=ls,label='sensitivity_level = {0:.2E}'.format(sensitivity_level))
            ax.set_ylabel('sensitivity',fontsize=fontsize)
            ax.legend()
            ax.grid()
        else:
            pass

        string = 'pixX - {0}'.format(xref)
        ax.set_xlabel(string,fontsize=fontsize)

        fig.tight_layout()
        if save:
            string = './{0}/{1}_spc1d.{2}'.format(savefolder,saveprefix,saveformat)
            fig.savefig(string,format=saveformat,bbox_inches='tight')
            print('Save {0}'.format(string))
